@@ -60,6 +60,7 @@ import univention.debug as ud_c
 import univention.debug2 as ud
 import univention.uldap
 from univention.connector.adcache import ADCache
+from univention.dn import DN
 
 
 term_signal_caught = False
@@ -278,6 +279,36 @@ class configdb(object):
                     self._dbcon.close()
                 self._dbcon = lite.connect(self.filename)
 
+    def get_rows_with_key_ending_in_dn(self, section, dn):
+        ending = '%%,%s' % dn
+        for _i in [1, 2]:
+            try:
+                cur = self._dbcon.cursor()
+                cur.execute("SELECT * FROM '%s' WHERE key LIKE ?" % section, (ending,))  # noqa: S608
+                rows = cur.fetchall()
+                cur.close()
+                return rows or []
+            except lite.Error as e:
+                ud.debug(ud.LDAP, ud.WARN, "sqlite: %s" % e)
+                if self._dbcon:
+                    self._dbcon.close()
+                self._dbcon = lite.connect(self.filename)
+
+    def get_rows_with_value_ending_in_dn(self, section, dn):
+        ending = '%%,%s' % dn
+        for _i in [1, 2]:
+            try:
+                cur = self._dbcon.cursor()
+                cur.execute("SELECT * FROM '%s' WHERE value LIKE ?" % section, (ending,))  # noqa: S608
+                rows = cur.fetchall()
+                cur.close()
+                return rows or []
+            except lite.Error as e:
+                ud.debug(ud.LDAP, ud.WARN, "sqlite: %s" % e)
+                if self._dbcon:
+                    self._dbcon.close()
+                self._dbcon = lite.connect(self.filename)
+
 
 class RFC4514_dn(object):
     special_dn_chars = '"+,;<=>'
@@ -514,6 +545,7 @@ class ucs(object):
 
         irrelevant_attributes = self.configRegistry.get('%s/ad/mapping/attributes/irrelevant' % (self.CONFIGBASENAME,), '')
         self.irrelevant_attributes = set(irrelevant_attributes.split(','))
+        self.allow_subtree_ancestors = self.configRegistry.is_true('%s/ad/mapping/allow-subtree-ancestors' % (self.CONFIGBASENAME,), False)
 
         # TODO just for testing, remove before merge
         # specific debug levels
@@ -682,6 +714,22 @@ class ucs(object):
         dn = self._get_dn_by_con(dn_con)
         return self.dn_mapped_to_base(dn, self.lo.base)
 
+    def _update_subtree_dns_in_mappings_by_con(self, old_con_dn, old_ucs_dn, new_con_dn, new_ucs_dn):
+        old_con_dn_lower = old_con_dn.lower()
+        old_ucs_dn_lower = old_ucs_dn.lower()
+        new_con_dn_lower = new_con_dn.lower()
+        new_ucs_dn_lower = new_ucs_dn.lower()
+        for row in self.config.get_rows_with_key_ending_in_dn('DN Mapping CON', old_con_dn_lower):
+            _new_key = self._subtree_replace(row[0], old_con_dn_lower, new_con_dn_lower)
+            _new_val = self._subtree_replace(row[1], old_ucs_dn_lower, new_ucs_dn_lower)
+            self._remove_config_option('DN Mapping CON', row[0])
+            self._set_config_option('DN Mapping CON', _new_key, _new_val)
+        for row in self.config.get_rows_with_value_ending_in_dn('DN Mapping UCS', old_con_dn_lower):
+            _new_key = self._subtree_replace(row[0], old_ucs_dn_lower, new_ucs_dn_lower)
+            _new_val = self._subtree_replace(row[1], old_con_dn_lower, new_con_dn_lower)
+            self._remove_config_option('DN Mapping UCS', row[0])
+            self._set_config_option('DN Mapping UCS', _new_key, _new_val)
+
     def _check_dn_mapping(self, dn_ucs, dn_con):
         dn_con_mapped = self._get_dn_by_ucs(dn_ucs.lower())
         dn_ucs_mapped = self._get_dn_by_con(dn_con.lower())
@@ -870,7 +918,7 @@ class ucs(object):
 
     def get_ucs_ldap_object_dn(self, dn):
         try:
-            return self.lo.lo.lo.search_s(dn, ldap.SCOPE_BASE, '(objectClass=*)', ('dn',))[0][0]
+            return self.lo.lo.lo.search_s(dn, ldap.SCOPE_BASE, '(objectClass=*)', ('1.1',))[0][0]
         except ldap.NO_SUCH_OBJECT:
             return
         except ldap.INVALID_DN_SYNTAX:
@@ -1231,7 +1279,6 @@ class ucs(object):
                     if isinstance(entryCSN, bytes):
                         entryCSN = entryCSN.decode('ASCII')
                     self._remember_entryCSN_commited_by_connector(entryUUID, entryCSN)
-            res = True
         return res
 
     def move_in_ucs(self, property_type, object, module, position):
@@ -1242,8 +1289,8 @@ class ucs(object):
         ud.debug(ud.LDAP, ud.INFO, "move_in_ucs: move object from %r to %r" % (object['olddn'], object['dn']))
         ucs_object = univention.admin.objects.get(module, None, self.lo, dn=object['olddn'], position='')
         ucs_object.open()
-        ucs_object.move(object['dn'])
-        return True
+        res = ucs_object.move(object['dn'])
+        return res
 
     def _get_entryUUID(self, dn):
         try:
@@ -1336,15 +1383,15 @@ class ucs(object):
         :param property_type:
                 the type of the object to be synced, must be part of the mapping. (e.g. "user", "group", "dc", "windowscomputer", etc.)
         :param object:
-                A dictionary describing the AD object.
+                A dictionary describing the AD object as returned by _object_mapping()
                 modtype: A modification type ("add", "modify", "move", "delete")
                 dn: The DN of the object in the UCS-LDAP
                 olddn: The olddn of the object object in UCS-LDAP (e.g. on "move" operation)
         :ptype object: dict
         :param pre_mapped_ad_dn:
-                pass
+                same as original_object['dn']
         :param original_object:
-                pass
+                as returned by __object_from_element()
         """
         # NOTE: pre_mapped_ad_dn means: original ad_dn (i.e. before _object_mapping)
         # this function gets an object from the ad class, which should be converted into a ucs module
@@ -1409,12 +1456,16 @@ class ucs(object):
 
             old_object = self.get_ucs_object(property_type, object.get('olddn', object['dn']))
 
+            # Corrections in case the target situation is unexpected:
             if old_object and object['modtype'] == 'add':
                 object['modtype'] = 'modify'
             if not old_object and object['modtype'] == 'modify':
                 object['modtype'] = 'add'
             if not old_object and object['modtype'] == 'move':
-                object['modtype'] = 'add'
+                if self.get_ucs_ldap_object_dn(object['dn']):
+                    object['modtype'] = 'modify'
+                else:
+                    object['modtype'] = 'add'
 
             if self.group_member_mapping_cache_ucs.get(object['dn'].lower()) and object['modtype'] != 'delete':
                 self.group_member_mapping_cache_ucs[object['dn'].lower()] = None
@@ -1444,14 +1495,42 @@ class ucs(object):
                 self._remove_dn_mapping(object['dn'], pre_mapped_ad_dn)
                 self.adcache.remove_entry(guid)
             if object['modtype'] == 'move':
-                result = self.move_in_ucs(property_type, object, module, position)
-                self._remove_dn_mapping(object['olddn'], '')  # we don't know the old ad-dn here anymore, will be checked by remove_dn_mapping
-                self._check_dn_mapping(object['dn'], pre_mapped_ad_dn)
+                res = self.move_in_ucs(property_type, object, module, position)
+                if isinstance(res, str):
+                    old_con_dn = original_object.get('olddn')
+                    if old_con_dn and pre_mapped_ad_dn != old_con_dn:
+                        self._remove_dn_mapping(object['olddn'], old_con_dn)
+                        self._update_group_related_caches(
+                            property_type,
+                            old_con_dn=old_con_dn,
+                            old_ucs_dn=object['olddn'],
+                            new_con_dn=pre_mapped_ad_dn,
+                            new_ucs_dn=object['dn'],
+                        )
+                    # self._remove_dn_mapping(object['olddn'], '')  # we don't know the old ad-dn here anymore, will be checked by remove_dn_mapping
+                    self._check_dn_mapping(object['dn'], pre_mapped_ad_dn)
+                    result = True
 
             if object['modtype'] == 'modify':
-                result = self.modify_in_ucs(property_type, object, module, position)
+                res = self.modify_in_ucs(property_type, object, module, position)
+                if res:
+                    if object['dn'].lower() != res.lower():  # This "modify" resulted in a modrdn in UDM even though the DN didn't change in AD
+                        self._remove_dn_mapping(object['dn'], pre_mapped_ad_dn)
+                        object['dn'] = res
+                    old_con_dn = original_object.get('olddn')
+                    if old_con_dn and pre_mapped_ad_dn != old_con_dn:
+                        self._remove_dn_mapping(object['olddn'], old_con_dn)
+                        self._update_group_related_caches(
+                            property_type,
+                            old_con_dn=old_con_dn,
+                            old_ucs_dn=object['olddn'],
+                            new_con_dn=pre_mapped_ad_dn,
+                            new_ucs_dn=object['dn'],
+                        )
+                # Finally commit the current DNs to the DN mapping cache
                 self._check_dn_mapping(object['dn'], pre_mapped_ad_dn)
                 self.adcache.add_entry(guid, original_object.get('attributes'))
+                result = True if res else res
 
             if not result:
                 ud.debug(ud.LDAP, ud.WARN, "Failed to get Result for DN (%r)" % (object['dn'],))
@@ -1484,14 +1563,14 @@ class ucs(object):
         return ldap.dn.str2dn(dn.lower())[-len(x):] == x
 
     @staticmethod
-    def _subtree_replace(dn, subtree, subtreereplace):
+    def _subtree_replace(dn, subtree, subtreereplace, case_folding=True):
         extra = ''
         if subtree.startswith(',') and subtreereplace.startswith(','):
             subtreereplace = subtreereplace[1:]
             subtree = subtree[1:]
             extra = ','
-        _dn = ldap.dn.str2dn(dn.lower())
-        _subtree = ldap.dn.str2dn(subtree.lower())
+        _dn = ldap.dn.str2dn(dn.lower()) if case_folding else ldap.dn.str2dn(dn)
+        _subtree = ldap.dn.str2dn(subtree.lower()) if case_folding else ldap.dn.str2dn(subtree)
         if _dn[-len(_subtree):] != _subtree or (extra and _dn == _subtree):
             return dn
         return ldap.dn.dn2str(ldap.dn.str2dn(dn)[:-len(_subtree)] + ldap.dn.str2dn(subtreereplace))
@@ -1634,10 +1713,10 @@ class ucs(object):
             ud.debug(ud.LDAP, dl, f"_ignore_object: ignore object without DN (key: {key})")
             return True  # ignore not existing object
 
-        if self.property[key].allow_subtree:
-            if not any(self._subtree_match(object['dn'], dn) for dn in self.property[key].allow_subtree):
-                ud.debug(ud.LDAP, dl, "_ignore_object: ignore object because it is not in one of the allowed subtrees: [%r:%r]" % (key, object['dn']))
-                return True
+        if self.property[key].allow_subtree and not any(self._subtree_match(object['dn'], dn) for dn in self.property[key].allow_subtree) and (
+                not self.allow_subtree_ancestors or not any(DN(subtree_dn).endswith(object['dn']) for subtree_dn in self.property[key].allow_subtree)):
+            ud.debug(ud.LDAP, dl, '_ignore_object: ignore object because it is not in one of the allowed subtrees: [%r:%r]' % (key, object['dn']))
+            return True
 
         if self.property[key].allow_filter:
             if not self._filter_match(self.property[key].allow_filter, object['attributes']):
